/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.task;

import com.vividsolutions.jts.geom.Envelope;
import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.geoviewer.Geographical;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.ItemIdType;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;
import org.openide.util.Lookup;

/**
 * Class which describes the ability to export a task
 * @author Maciek Dulkiewicz <maciek.dulkiewicz@ordina.nl>
 */
public abstract class ExportableTask extends Task {

    public static final String PROP_EXPORT_PREVIEW_GRID = "grid";
    public static final String PROP_WORK_WIDTH = "workWidth";
    public static final String PROP_AVERAGE_TZN_VALUE = "averageTznValue";
    public static final String PROP_DEFAULT_TZN_VALUE = "defaultTznValue";
    public static final String MIME_TYPE = "application/x-agrosense-exportable-task";
    public static final String PROP_REGENERATE_GRID = "regenerateGrid";

    private ExportableTaskCapability exportableTaskCapability = new ExportableTaskCapability(this);
    private transient GridCoverage2D exportPreviewGrid;
    private transient boolean regenerateGrid = true;
    // track last modification timestamp for the above flag, as new changes may have occured during generation:
    private transient volatile long regenerateGridTimestamp;
    private Double averageTznValue; // read-only, calculated (might not need field to cache avg val), should auto-update on tzn.value changes
    private Double defaultTznValue; // editable, out of field / loss of GPS rate, default to average when not set by user

    public ExportableTask(EnterpriseIdProvider eip, ItemIdType itemIdType) {
        this(eip, new Class[0], itemIdType);
    }

    public ExportableTask(EnterpriseIdProvider eip, Class[] requires, ItemIdType itemIdType) {
        super(eip, requires, itemIdType);
    }
    
    public ActivityField getActivityField(){
        return getLookup().lookup(ActivityField.class);
    }

    public abstract double getWorkWidth();

    @Override
    protected void setStatus(Status status) {
        super.setStatus(status);

        if (Status.COMPLETE.equals(status) && getLookup().lookup(ExportableTaskCapability.class) == null) {
            ic.add(exportableTaskCapability);
        } else if (Status.INCOMPLETE.equals(status) && getLookup().lookup(ExportableTaskCapability.class) != null) {
            ic.remove(exportableTaskCapability);
        }

    }

    /**
     * <p>Validates the exportable task</p>
     * <p>Will validate if there are no empty {@link TreatmentZone#value}
     * in the {@link TreatmentZone} 's used to create the {@link TreatmentZoneGrid}
     * </p>
     * 
     * @return Array of ExportableTaskErrorEnum's for every validation issue that is found
     */
    public ExportableTaskError[] validate() {
        // TODO: 
        // only return an error when all values are null or no tzn present? (see {@code hasValues()})
        // if only some values are null a preview can still be shown, and the export will use the out-of-field value;
        // we could still show a warning in these cases;
        // only if there aren't any values at all the preview and export won't have 
        // any useful data, and in that case the validation should also be used to block the export
        // (currently it only blocks the preview).
        ActivityField activityField = getLookup().lookup(ActivityField.class);
        List<TreatmentZone> treatmentZones = activityField.getTreatmentZones();
        for(TreatmentZone treatmentZone : treatmentZones){
            if(treatmentZone.getValue() == null){
                return new ExportableTaskError[]{ExportableTaskError.TREATMENT_ZONE_NULL_VALUE};
            }
        }
        return new ExportableTaskError[0];
    }
    
    /**
     * Checks if there are any non-null treatment zone values available.
     * <p>A more lenient check than {@link validate()}, to be used by grid previews.</p>
     * @return 
     */
    public boolean hasTreatmentZoneValues() {
        ActivityField activityField = getActivityField();
        if (activityField == null) { return false; }
        List<TreatmentZone> treatmentZones = activityField.getTreatmentZones();
        if (treatmentZones == null) { return false; }
        for(TreatmentZone treatmentZone : treatmentZones){
            if(treatmentZone.getValue() != null) { return true; }
        }
        return false;
    }

    /**
     * retrieve a grid with the preview of what the task looks like.
     * 
     * Warning: if grid is not initialized, this method will invoke synchronous creation
     * of the preview grid, which might take a while.
     * 
     * @return GridCoverage2D grid preview
     */
    public GridCoverage2D getExportPreviewGrid(){
        if (regenerateGrid()) {
            if (exportPreviewGrid == null) {
                exportPreviewGrid = getDummyRaster();
            }
            final long timestamp = regenerateGridTimestamp;
            new SwingWorker<Object, Object>() {
                
                @Override
                protected Object doInBackground() throws Exception {
                    createExportableGrid();
                    return null;
                }

                @Override
                protected void done() {
                    if (timestamp == regenerateGridTimestamp) {
                        // only if no new changes have occured during generation:
                        setRegenerateGrid(false);
                    }
                }
                
            }.execute();
        }
        return exportPreviewGrid;
    }

    public boolean regenerateGrid() {
        return regenerateGrid;
    }

    public void setRegenerateGrid(boolean regenerate) {
        boolean old = regenerateGrid;
        regenerateGrid = regenerate;
        getPropertyChangeSupport().firePropertyChange(PROP_REGENERATE_GRID, old, regenerate);
        
    }

    /**
     * This method creates and returns a {@link TreatmentZoneGrid} using the
     * treatment zones contained by the {@link TaskContext} of this task. The
     * number of treatment zones used in the TreatmentZoneGrid is limited by the
     * maximum defined at the top of this class.
     *
     * @return A TreatmentZoneGrid containing the maximum number of treatment
     * zones supported by the current ISOBUS implementation
     */
    private void createExportableGrid(){
        //TODO what to do with multiple implementations?
        TaskExporter exporter = Lookup.getDefault().lookup(TaskExporter.class);
        final GridCoverage2D old = this.exportPreviewGrid;
        this.exportPreviewGrid = exporter.generateExportPreviewGrid(this);
        
        SwingUtilities.invokeLater(new Runnable() {

            @Override
            public void run() {
                getPropertyChangeSupport().firePropertyChange(PROP_EXPORT_PREVIEW_GRID, old, ExportableTask.this.exportPreviewGrid);
            }
        });
    }


    /**
     * Calculate the average of the distinct treatment zone values in the tasks.
     * 
     * @param tasks
     * @return the average, or null when no values are available.
     */
    public static Double getAverageTznValue(final ExportableTask... tasks) {
        // maybe later: different strategies for average,
        // e.g. distinct values, all values, weighted over area, ...
        if (tasks == null) { return null; }
        
        final Set<Double> values = new TreeSet<>();
        for (ExportableTask task : tasks) {
            ActivityField afd = task.getActivityField();
            for (TreatmentZone tzn : afd.getTreatmentZones()) {
                if (tzn.getValue() != null) {
                    values.add(tzn.getValue());
                }
            }
        }
        
        if (values.isEmpty()) { return null; }

        double sum = 0.0;
        for (Double value : values) {
            sum += value;
        }
        //https://issuetracker.kvg-mechatronics.nl/bug_view_page.php?bug_id=4815
        // round to 1 digit
        return Math.round((sum / values.size())*10)/10.0d;
    }

    public Double getDefaultTznValue() {
        return defaultTznValue;
    }

    public void setDefaultTznValue(Double defaultTznValue) {
        Double old = this.defaultTznValue;
        this.defaultTznValue = defaultTznValue;
        getPropertyChangeSupport().firePropertyChange(PROP_DEFAULT_TZN_VALUE, old, this.defaultTznValue);
    }

    public Double getAverageTznValue() {
        return averageTznValue;
    }
    
    public void recalculateAverageTznValue() {
        Double old = averageTznValue;
        averageTznValue = getAverageTznValue(this);
        getPropertyChangeSupport().firePropertyChange(PROP_AVERAGE_TZN_VALUE, old, averageTznValue);
    }
    
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        regenerateGrid = true;
    }
    
    public  GridCoverage2D getDummyRaster() {
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, 1,1,1, new Point(0,0));
        raster.setSample(0,0,0,Integer.MIN_VALUE);
        Envelope e = getActivityField().getBoundingBox().getEnvelopeInternal();
        BoundingBox boundingBox = new ReferencedEnvelope(e.getMinX(), e.getMaxX(), e.getMinY(), e.getMaxY(), DefaultGeographicCRS.WGS84);
        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);        
    }
    
    public abstract class ExportableTaskGeographical implements Geographical<GridCoverage2D>, Geographical.AsyncRenderObject, Serializable {
        private PropertyChangeSupport pcs;

        @Override
        public GridCoverage2D getRenderObject(Envelope envelope) {
            return ExportableTask.this.getExportPreviewGrid();
        }
        
        protected final PropertyChangeSupport getPropertyChangeSupport() {
            if (pcs == null) {
                pcs = new PropertyChangeSupport(this);
            }
            return pcs;
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener listener) {
            getPropertyChangeSupport().addPropertyChangeListener(listener);
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener listener) {
            getPropertyChangeSupport().removePropertyChangeListener(listener);
        }

        @Override
        public boolean hasStaleRenderObject() {
            return ExportableTask.this.regenerateGrid();
        }
    }
}
